/*
 * Wn -> WGS84 ϊ
 */
/* $Id: wgstko.c,v 1.7 2001/12/02 13:26:23 tmurakam Exp $ */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "wgstko.h"

#define	DATUM_WGS84	0
#define	DATUM_TOKYO	1

#define	a	6377397.155
#define	a84	6378137.0
#define	f84	0.0033528106647478
#define	f_tokyo	0.00334277318
#define	PI	3.141592653589793

/*
 * ܓxEoxEx𒼌Wɕϊ
 */
void dms_to_XYZ(double lat, double lon, double height,
		double *X, double *Y, double *Z, int datum)
{
	double e_square, N, f, adms;

	if (datum == DATUM_WGS84) {
		adms = a84;
		f = f84;
	} else {
		adms = a;
		f = f_tokyo;
	}

	e_square = f * (2 - f);
	N = adms / sqrt(1 - e_square * sin(lat) * sin(lat));
	*X = (N + height) * cos(lat) * cos(lon);
	*Y = (N + height) * cos(lat) * sin(lon);
	*Z = (N * (1 - e_square) + height) * sin(lat);
}

/*
 * WܓxEoxExɕϊ
 */
void XYZ_to_dms(double X, double Y, double Z, 
		double *lat, double *lon, double *height, int datum)
{
	double b_dash, P, theta, E_square, E_square_dash, neu, f, adms;

	if (datum == DATUM_WGS84) {
		adms = a84;
		f = f84;
	} else {
		adms = a;
		f = f_tokyo;
	}

	b_dash = adms * (1 - f);
	P = sqrt(X*X + Y*Y);
	theta = atan((Z * adms) / (P * b_dash));
	E_square = (adms * adms - b_dash * b_dash) / (adms * adms);
	E_square_dash = (adms * adms - b_dash * b_dash) / (b_dash * b_dash);

	*lat = atan((Z + E_square_dash*b_dash*sin(theta)*sin(theta)*sin(theta))/
		    (P - E_square * adms * cos(theta)*cos(theta)*cos(theta)));
	*lon = atan(Y/X) + PI;
	neu = adms / sqrt(1 - E_square * sin(*lat) * sin(*lat));
	*height = P/cos(*lat) - neu;
}

/*
 * p[^ϊ
 */
void convert_param(double *X, double *Y, double *Z, int datum)
{
	if (datum == DATUM_WGS84) {
		*X += -147.54;
		*Y += 507.26;
		*Z += 680.47;
	} else {
		*X += 147.54;
		*Y += -507.26;
		*Z += -680.47;
	}
}

#define	deg2rad(x)	((x) * PI / 180.0)
#define	rad2deg(x)	((x) * 180.0 / PI)

/*
 * Wϊ
 * xp[^͂ł͖B
 */
void tko2wgs(double *plat, double *plon)
{
	double lat, lon, height;
	double X, Y, Z;

	lat = deg2rad(*plat);
	lon = deg2rad(*plon);

	dms_to_XYZ(lat, lon, 0, &X, &Y, &Z, DATUM_TOKYO);
	convert_param(&X, &Y, &Z, DATUM_WGS84);
	XYZ_to_dms(X, Y, Z, &lat, &lon, &height, DATUM_WGS84);
	
	*plat = rad2deg(lat);
	*plon = rad2deg(lon);
}

